local PI = math.pi

local function INT(number) return math.floor(number) end

local function jdFromDate(dd, mm, yy)
    local a, y, m, jd;
    a = INT((14 - mm) / 12);
    y = yy + 4800 - a;
    m = mm + 12 * a - 3;
    jd = dd + INT((153 * m + 2) / 5) + 365 * y + INT(y / 4) - INT(y / 100) +
             INT(y / 400) - 32045;
    if (jd < 2299161) then
        jd = dd + INT((153 * m + 2) / 5) + 365 * y + INT(y / 4) - 32083;
    end
    return jd
end

local function jdToDate(jd)
    local a, b, c, d, e, m, day, month, year, o;
    if (jd > 2299160) then
        a = jd + 32044;
        b = INT((4 * a + 3) / 146097);
        c = a - INT((b * 146097) / 4);
    else
        b = 0;
        c = jd + 32082;
    end
    d = INT((4 * c + 3) / 1461);
    e = c - INT((1461 * d) / 4);
    m = INT((5 * e + 2) / 153);
    day = e - INT((153 * m + 2) / 5) + 1;
    month = m + 3 - 12 * INT(m / 10);
    year = b * 100 + d - 4800 + INT(m / 10);
    o = day + '/' + month + '/' + year
    return o
end

local function getNewMoonDay(k, timeZone)
    local T, T2, T3, dr, Jd1, M, Mpr, F, C1, deltat, JdNew;
    T = k / 1236.85;
    T2 = T * T;
    T3 = T2 * T;
    dr = PI / 180;
    Jd1 = 2415020.75933 + 29.53058868 * k + 0.0001178 * T2 - 0.000000155 * T3;
    Jd1 = Jd1 + 0.00033 * math.sin((166.56 + 132.87 * T - 0.009173 * T2) * dr);
    M = 359.2242 + 29.10535608 * k - 0.0000333 * T2 - 0.00000347 * T3;
    Mpr = 306.0253 + 385.81691806 * k + 0.0107306 * T2 + 0.00001236 * T3;
    F = 21.2964 + 390.67050646 * k - 0.0016528 * T2 - 0.00000239 * T3;
    C1 = (0.1734 - 0.000393 * T) * math.sin(M * dr) + 0.0021 *
             math.sin(2 * dr * M);
    C1 = C1 - 0.4068 * math.sin(Mpr * dr) + 0.0161 * math.sin(dr * 2 * Mpr);
    C1 = C1 - 0.0004 * math.sin(dr * 3 * Mpr);
    C1 = C1 + 0.0104 * math.sin(dr * 2 * F) - 0.0051 * math.sin(dr * (M + Mpr));
    C1 = C1 - 0.0074 * math.sin(dr * (M - Mpr)) + 0.0004 *
             math.sin(dr * (2 * F + M));
    C1 = C1 - 0.0004 * math.sin(dr * (2 * F - M)) - 0.0006 *
             math.sin(dr * (2 * F + Mpr));
    C1 = C1 + 0.0010 * math.sin(dr * (2 * F - Mpr)) + 0.0005 *
             math.sin(dr * (2 * Mpr + M));
    if (T < -11) then
        deltat = 0.001 + 0.000839 * T + 0.0002261 * T2 - 0.00000845 * T3 -
                     0.000000081 * T * T3;
    else
        deltat = -0.000278 + 0.000265 * T + 0.000262 * T2;
    end
    JdNew = Jd1 + C1 - deltat;
    return INT(JdNew + 0.5 + timeZone / 24)
end

local function getSunLongitude(jdn, timeZone)
    local T, T2, dr, M, L0, DL, L;
    T = (jdn - 2451545.5 - timeZone / 24) / 36525;
    T2 = T * T;
    dr = PI / 180;
    M = 357.52910 + 35999.05030 * T - 0.0001559 * T2 - 0.00000048 * T * T2;
    L0 = 280.46645 + 36000.76983 * T + 0.0003032 * T2;
    DL = (1.914600 - 0.004817 * T - 0.000014 * T2) * math.sin(dr * M);
    DL = DL + (0.019993 - 0.000101 * T) * math.sin(dr * 2 * M) + 0.000290 *
             math.sin(dr * 3 * M);
    L = L0 + DL;
    L = L * dr;
    L = L - PI * 2 * (INT(L / (PI * 2)));
    return INT(L / PI * 6)
end

local function getLunarMonth11(yy, timeZone)
    local k, off, nm, sunLong;
    off = jdFromDate(31, 12, yy) - 2415021;
    k = INT(off / 29.530588853);
    nm = getNewMoonDay(k, timeZone);
    sunLong = getSunLongitude(nm, timeZone);
    if (sunLong >= 9) then nm = getNewMoonDay(k - 1, timeZone); end
    return nm;
end

local function getLeapMonthOffset(a11, timeZone)
    local k, last, arc, i;
    k = INT((a11 - 2415021.076998695) / 29.530588853 + 0.5);
    last = 0;
    i = 1;
    arc = getSunLongitude(getNewMoonDay(k + i, timeZone), timeZone);
    while not (arc == last) and (i < 14) do
        last = arc;
        i = i + 1;
        arc = getSunLongitude(getNewMoonDay(k + i, timeZone), timeZone);
    end
    return i - 1;
end

function ConvertSolar2Lunar(dd, mm, yy, timeZone, addYear)
    local k, dayNumber, monthStart, a11, b11, lunarDay, lunarMonth, lunarYear,
          lunarLeap, o, diff, leapMonthDiff;
    dayNumber = jdFromDate(dd, mm, yy);
    k = INT((dayNumber - 2415021.076998695) / 29.530588853);
    monthStart = getNewMoonDay(k + 1, timeZone);
    if (monthStart > dayNumber) then monthStart = getNewMoonDay(k, timeZone); end
    a11 = getLunarMonth11(yy, timeZone);
    b11 = a11;
    if (a11 >= monthStart) then
        lunarYear = yy;
        a11 = getLunarMonth11(yy - 1, timeZone);
    else
        lunarYear = yy + 1;
        b11 = getLunarMonth11(yy + 1, timeZone);
    end
    lunarDay = dayNumber - monthStart + 1;
    diff = INT((monthStart - a11) / 29);
    lunarLeap = 0;
    lunarMonth = diff + 11;
    if (b11 - a11 > 365) then
        leapMonthDiff = getLeapMonthOffset(a11, timeZone);
        if (diff >= leapMonthDiff) then
            lunarMonth = diff + 10;
            if (diff == leapMonthDiff) then lunarLeap = 1; end
        end
    end
    if (lunarMonth > 12) then lunarMonth = lunarMonth - 12; end
    if (lunarMonth >= 11 and diff < 4) then lunarYear = lunarYear - 1; end
    if (lunarDay < 10) then
        o = '0'
    else
        o = ''
    end
    o = o .. lunarDay .. '/'
    if (lunarMonth < 10) then
        o = o .. '0'
    end
    o = o .. lunarMonth

    if addYear then
        o = o .. '/' .. lunarYear
    end
    return o
end

function ConvertLunar2Solar(lunarDay, lunarMonth, lunarYear, lunarLeap, timeZone)
    local k, a11, b11, off, leapOff, leapMonth, monthStart;
    if (lunarMonth < 11) then
        a11 = getLunarMonth11(lunarYear - 1, timeZone);
        b11 = getLunarMonth11(lunarYear, timeZone);
    else
        a11 = getLunarMonth11(lunarYear, timeZone);
        b11 = getLunarMonth11(lunarYear + 1, timeZone);
    end
    off = lunarMonth - 11;
    if (off < 0) then off = off + 12; end
    if (b11 - a11 > 365) then
        leapOff = getLeapMonthOffset(a11, timeZone);
        leapMonth = leapOff - 2;
        if (leapMonth < 0) then off = off + 12; end
        if (not (lunarLeap == 0) and not (lunarMonth == leapMonth)) then
            return '0/0/0'
        elseif (not (lunarLeap == 0) or (off >= leapOff)) then
            off = off + 1;
        end
    end
    k = INT(0.5 + (a11 - 2415021.076998695) / 29.530588853);
    monthStart = getNewMoonDay(k + off, timeZone);
    return jdToDate(monthStart + lunarDay - 1);
end
